using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Linq.Expressions;
using Baseline;
using Marten.Schema.Identity;
using Marten.Util;
using Npgsql;

namespace Marten.Schema.BulkLoading
{
    public class BulkLoader<T> : IBulkLoader<T>
    {
        private readonly IdAssignment<T> _assignment;
        private readonly string _baseSql;
        private readonly DocumentMapping _mapping;
        private readonly string _sql;

        private readonly Action<T, string, ISerializer, NpgsqlBinaryImporter> _transferData;
        private readonly string _tempTableName;


        public BulkLoader(ISerializer serializer, DocumentMapping mapping, IdAssignment<T> assignment)
        {
            _mapping = mapping;
            _assignment = assignment;
            var upsertFunction = new UpsertFunction(mapping);

            _tempTableName = mapping.Table.Name + "_temp";


            var writer = Expression.Parameter(typeof(NpgsqlBinaryImporter), "writer");
            var document = Expression.Parameter(typeof(T), "document");
            var alias = Expression.Parameter(typeof(string), "alias");
            var serializerParam = Expression.Parameter(typeof(ISerializer), "serializer");

            var arguments = upsertFunction.OrderedArguments().ToArray();
            var expressions =
                arguments.Select(
                    x => x.CompileBulkImporter(serializer.EnumStorage, writer, document, alias, serializerParam));

            var columns = arguments.Select(x => $"\"{x.Column}\"").Join(", ");
            _baseSql = $"COPY %TABLE%({columns}) FROM STDIN BINARY";
            _sql = _baseSql.Replace("%TABLE%", mapping.Table.QualifiedName);

            var block = Expression.Block(expressions);

            var lambda = Expression.Lambda<Action<T, string, ISerializer, NpgsqlBinaryImporter>>(block, document, alias,
                serializerParam, writer);

            _transferData = lambda.Compile();
        }

        public void Load(ISerializer serializer, NpgsqlConnection conn, IEnumerable<T> documents)
        {
            load(serializer, conn, documents, _sql);
        }

        public void Load(TableName table, ISerializer serializer, NpgsqlConnection conn, IEnumerable<T> documents)
        {
            var sql = _baseSql.Replace("%TABLE%", table.QualifiedName);
            load(serializer, conn, documents, sql);
        }

        public void LoadIntoTempTable(ISerializer serializer, NpgsqlConnection conn, IEnumerable<T> documents)
        {
            var sql = _baseSql.Replace("%TABLE%", _tempTableName);
            load(serializer, conn, documents, sql);
        }

        public string CopyNewDocumentsFromTempTable()
        {
            var table = _mapping.SchemaObjects.StorageTable();

            var storageTable = table.Table.QualifiedName;
            var columns = table.Columns.Select(x => $"\"{x.Name}\"").Join(", ");
            var selectColumns = table.Columns.Select(x => $"{_tempTableName}.\"{x.Name}\"").Join(", ");

            return $@"insert into {storageTable} ({columns}) (select {selectColumns} from {_tempTableName} 
                         left join {storageTable} on {_tempTableName}.id = {storageTable}.id where {storageTable}.id is null)";
        }

        public string OverwriteDuplicatesFromTempTable()
        {
            var table = _mapping.SchemaObjects.StorageTable();
            var storageTable = table.Table.QualifiedName;

            var updates = table.Columns.Where(x => x.Name != "id")
                .Select(x => $"{x.Name} = source.{x.Name}").Join(", ");

            return $@"update {storageTable} target SET {updates} FROM {_tempTableName} source WHERE source.id = target.id";
        }

        public string CreateTempTableForCopying()
        {
            var tempTable = StorageTable.Name + "_temp";

            return $"create temporary table {tempTable} as select * from {StorageTable.QualifiedName};truncate table {_tempTableName}";
        }

        public TableName StorageTable => _mapping.Table;

        private void load(ISerializer serializer, NpgsqlConnection conn, IEnumerable<T> documents, string sql)
        {
            using (var writer = conn.BeginBinaryImport(sql))
            {
                foreach (var document in documents)
                {
                    var assigned = false;
                    _assignment.Assign(document, out assigned);

                    writer.StartRow();

                    _transferData(document, _mapping.AliasFor(document.GetType()), serializer, writer);
                }
            }
        }
    }
}
 